unit GrTools;

interface

uses WinTypes, WinProcs, Objects, OWindows;

type
    ColorData = record
      DC: HDC;
      Pen: HPen;
      Brush: HBrush;
      Reseted,
      StockPen,
      StockBrush: Boolean;
    end;

    PInfoWindow = ^TInfoWindow;
    TInfoWindow = Object(TWindow)
      pInfo: array[0..255] of char;
      InfoLen: Integer;
      constructor Init(AParent: PWindowsObject; InfoText: String);
      procedure   Paint(PaintDC: HDC; var PaintInfo: TPaintStruct); virtual;
  end;

var InfoPtr: PInfoWindow;
    

function Imin(X, Y: Integer): Integer;
function Imax(X, Y: Integer): Integer;
function min(X, Y: Real): Real;
function max(X, Y: Real): Real;
procedure InvertLine(DC: HDC; X1, Y1, X2, Y2: Integer; Muster: Word);

procedure GiveWindowTime;
procedure GiveWindowTimeUndisable;
procedure EnableWindowTime;
procedure DisableWindowTime;

procedure Message(S: String);
function  MessageString(WndParent: HWnd; Txt, Caption: String; TextType: Word): Integer;
procedure AppMessage(S: String);
function  AppMessageString(Txt, Caption: String; TextType: Word): Integer;

procedure ShowInfo(S: String);
procedure HideInfo;

function NToS(X: Real): String;
function LTrim(s: String): String;
function RTrim(s: String): String;
function AllTrim(s: String): String;

function StrPCat(Dest: PChar; Source: String): PChar;

procedure SetColor(var SaveColor: ColorData; ADC: HDC; PCol, BCol: LongInt);
procedure ResetColor(SaveColor: ColorData); 

function CMYK(c, m, y, k: Real): LongInt;

implementation

uses Strings;

const
    WindowTimeEnabled: Boolean = True;


procedure EnableWindowTime;
begin
  WindowTimeEnabled := True;
end;

procedure DisableWindowTime;
begin
  WindowTimeEnabled := False;
end;

procedure GiveWindowTimeUndisable;
 var
    MyMsg: TMsg;
begin
  IF PeekMessage(MyMsg, 0, 0, 0, pm_Remove) THEN
  begin
      TranslateMessage(MyMsg);
      DispatchMessage(MyMsg);
  end;
end;

procedure GiveWindowTime;
 var
    MyMsg: TMsg;
begin
  if WindowTimeEnabled then
    GiveWindowTimeUndisable;
end;

Procedure Message(s: String);
begin
  MessageString(0, s, '', mb_ok);
end;

function MessageString(WndParent: HWnd; Txt, Caption: String; TextType: Word): Integer;
 var
    a, b: array[0..255] of char;
begin
  StrPCopy(a, Txt);
  StrPCopy(b, Caption);
  MessageString := MessageBox(WndParent, a, b, TextType);
end;

Procedure AppMessage(s: String);
begin
  MessageString(Application^.MainWindow^.HWindow, s, '', mb_ok);
end;

function AppMessageString(Txt, Caption: String; TextType: Word): Integer;
begin
  AppMessageString := MessageString(Application^.MainWindow^.HWindow, Txt, Caption, TextType);
end;


{----------------------- TInfoWindow ---------------------}

constructor TInfoWindow.Init(AParent: PWindowsObject; InfoText: String);
 var
    Extent: LongInt;
    X, Y: Integer;
    LocalDC: HDC;
begin
  TWindow.Init(AParent, nil);
  Attr.Style := ws_PopUp or ws_Visible;
  StrPCopy(pInfo, InfoText);
  InfoLen := Length(InfoText);

  LocalDC := GetDC(0);
  Extent := GetTextExtent(LocalDC, pInfo, InfoLen);
  X := GetDeviceCaps(LocalDC, HorzRes);
  Y := GetDeviceCaps(LocalDC, VertRes);
  ReleaseDC(0, LocalDC);
  Attr.W := LoWord(Extent)+20;
  Attr.H := HiWord(Extent)+20;
  Attr.X := (X-Attr.W) div 2;
  Attr.Y := (Y-Attr.H) div 2;

end;


procedure TInfoWindow.Paint(PaintDC: HDC; var PaintInfo: TPaintStruct);
var
  r: TRect;		
begin
{  TWindow.Paint(PaintDC, PaintInfo);}
  SetRect(r, 0, 0, Attr.W, Attr.H);
  FrameRect(PaintDC, r, GetStockObject(Black_Brush));
  SetRect(r, 1, 1, Attr.W-1, Attr.H-1);
  FillRect(PaintDC, r, GetStockObject(White_Brush));
  TextOut(PaintDC, 10, 10, pInfo, InfoLen);
end;


procedure ShowInfo(s: String);
begin
  if InfoPtr<>nil then
    Dispose(InfoPtr, Done);
  InfoPtr := New(PInfoWindow, Init(nil, s));
  InfoPtr^.Create;
  UpdateWindow(InfoPtr^.HWindow);
end;

procedure HideInfo;
begin
  if InfoPtr<>nil then
    Dispose(InfoPtr, Done);
  InfoPtr := nil;
end;



function NToS(X: Real): String;
 var
    li: LongInt;
    r: Real;
    s: String;
begin
  if X = int(X) then
  begin
    if (X<=MaxLongInt) and (X>=-MaxLongInt) then
      str(trunc(X), s)
    else
      str(x: 18: 0, s);
  end
  else
  begin
    str(x: 18: 4, s);
    while s[length(s)]='0' do
      s := copy(s,1,length(s)-1);
    if s[length(s)]='.' then
      s := copy(s,1,length(s)-1);
  end;
  NToS := LTrim(s);
end;

function LTrim(s: String): String;
 var
    i: Integer;
begin
  i := 1;
  while (i<length(s)) and (s[i]=' ') do
    i := i+1;
  LTrim := copy(s,i,length(s));
end;

function RTrim(s: String): String;
 var
    i: Integer;
begin
  i := length(s);
  while (i>0) and (s[i]=' ') do
    i := i-1;
  RTrim := copy(s,1, i);
end;

function AllTrim(s: String): String;
begin
  AllTrim := RTrim(LTrim(s));
end;

function StrPCat(Dest: PChar; Source: String): PChar;
 var
    a: array[0..255] of char;
begin
  StrPCopy(a, Source);
  StrCat(Dest, a);
  StrPCat := Dest;
end;



{
uses strings;

type
    PData = ^DataType;
    DataType = Record
      DC: HDC;
      Count: Integer;
    end;

var
   Data: DataType;
procedure InvertPixel(X, Y: Integer; DataPtr: PData); far;
 var
    r: TRect;
    a: array[0..80] of char;
    s1, s2, s3: String;
begin
  str(x,s1);
  str(y,s2);
  str(DataPtr^.Count, s3);
  strpcopy(a, 'X='+s1+'; Y='+s2+'; Count='+s3);

  MessageBox(0, a, 'IP', mb_ok);
  DataPtr^.Count := DataPtr^.Count+1;
  SetRect(r, X, Y, X+1, Y+1);
  InvertRect(DataPtr^.DC, r); 
end;

procedure InvertLine(DC: HDC; X1, Y1, X2, Y2: Integer);
begin
  Data.DC := DC;
  Data.Count := 0;
  LineDDA(X1, Y1, X2, Y2, @InvertPixel, @Data);
end;
}

const
    MinX =   0;
    MaxX = 320;
    MinY =   0;
    MaxY = 240;


function Imin(X, Y: Integer): Integer;
begin
  if X<Y then
    IMin := X
  else
    IMin := Y;
end;

function Imax(X, Y: Integer): Integer;
begin
  if X>Y then
    IMax := X
  else
    IMax := Y;
end;

function min(X, Y: Real): Real;
begin
  if X<Y then
    Min := X
  else
    Min := Y;
end;

function max(X, Y: Real): Real;
begin
  if X>Y then
    Max := X
  else
    Max := Y;
end;

procedure InvertLine(DC: HDC; X1, Y1, X2, Y2: Integer; Muster: Word);
 var
    r: TRect;
    i, dx, dy, sx, sy, xk, yk, X, Y, XS, YS,
    XMin, XMax, YMin, YMax: Integer;
    Maske: Word;

 procedure InvertPixel(X, Y: Integer);
 begin
   if (X<=MaxX) and (X>=MinX) and (Y<=MaxY) and (Y>=MinY) then
   begin
     SetRect(r, X, Y, X+1, Y+1);
     InvertRect(DC, r);
   end;
 end;

begin
  XMin := Imin(X1, X2);
  YMin := Imin(Y1, Y2);
  XMax := Imax(X1, X2);
  YMax := Imax(Y1, Y2);
  if (XMax>=MinX) and (XMin<=MaxX) and (YMax>=MinY) and (YMin<=MaxY) then
  begin
    if (XMax>MaxX) or (XMin<MinX) or (YMax>MaxY) or (YMin<MinY) then
    begin
      dx := X2-X1;
      dy := Y2-Y1;
      if (dx=0) or (dy=0) then
      begin
        X1 := Imax(XMin, MinX);
        Y1 := Imax(YMin, MinY);
        X2 := Imin(XMax, MaxX);
        Y2 := Imin(YMax, MaxY);
      end
      else
      begin
        if XMax>MaxX then
        begin
          XS := MaxX;
          YS := Y1+trunc(1.0*(XS-X1)*dy/dx);
          if XMax=X2 then
          begin
            X2 := XS;
            Y2 := YS;
          end
          else
          begin
            X1 := XS;
            Y1 := YS;
          end;
        end;
        if XMin<MinX then
        begin
          XS := MinX;
          YS := Y1+trunc(1.0*(XS-X1)*dy/dx);
          if XMin=X2 then
          begin
            X2 := XS;
            Y2 := YS;
          end
          else
          begin
            X1 := XS;
            Y1 := YS;
          end;
        end;
        YMin := Imin(Y1,Y2);
        YMax := Imax(Y1,Y2);
        if (YMin>=MaxY) and (YMax<=MinY) then
        begin
          if YMax>MaxY then
          begin
            YS := MaxY;
            XS := X1+trunc(1.0*(YS-Y1)*dx/dy);
            if YMax=Y2 then
            begin
              X2 := XS;
              Y2 := YS;
            end
            else
            begin
              X1 := XS;
              Y1 := YS;
            end;
          end;
          if YMin<MinY then
          begin
            YS := MinY;
            XS := X1+trunc(1.0*(YS-Y1)*dx/dy);
            if YMin=Y2 then
            begin
              X2 := XS;
              Y2 := YS;
            end
            else
            begin
              X1 := XS;
              Y1 := YS;
            end;
          end;
        end;
      end;
      XMin := Imin(X1, X2);
      YMin := Imin(Y1, Y2);
      XMax := Imax(X1, X2);
      YMax := Imax(Y1, Y2);
    end;

    if (XMax>=MinX) and (XMin<=MaxX) and (YMax>=MinY) and (YMin<=MaxY) then
    begin
      Maske := 1;
      dx := X2-X1;
      dy := Y2-Y1;
      if dx<0 then
        sx := -1
      else
        sx :=  1;
      if dy<0 then
        sy := -1
      else
        sy :=  1;
      dx := abs(dx);
      dy := abs(dy);

      if dx>dy then
        i := dx
      else
        i := dy;

      xk := dx div 2;
      yk := dy div 2;
      X := X1;
      Y := Y1;
      while (i>0) do
      begin
        if xk>yk then
        begin
          xk := xk-yk;
          yk := 0;
        end
        else
        begin
          yk := yk-xk;
          xk := 0;
        end;
        i := i-1;
        if (Muster and Maske)<>0 then
          InvertPixel(X, Y);
        if Maske=$8000 then
          Maske := 1
        else
          Maske := Maske shl 1;
        if xk=yk then
        begin
          Y := Y+sy;
          xk := xk+dx;
          X := X+sx;
          yk := yk+dy;
        end
        else if xk<yk then
        begin
          Y := Y+sy;
          xk := xk+dx;
          if xk>yk then
          begin
            X := X+sx;
            yk := yk+dy;
          end;
        end
        else if xk>yk then
        begin
          X := X+sx;
          yk := yk+dy;
          if xk<yk then
          begin
            Y := Y+sy;
            xk := xk+dx;
          end;
        end;
      end;
      InvertPixel(X2, Y2);
    end;
  end;
end;


procedure SetColor(var SaveColor: ColorData; ADC: HDC; PCol, BCol: LongInt);
begin
  With SaveColor do
  begin
    DC := ADC;
    StockPen := (PCol=-1);
    StockBrush := (BCol=-1);
    if StockPen then
      Pen := SelectObject(ADC, GetStockObject(Null_Pen))
    else
      Pen := SelectObject(ADC, CreatePen(ps_Solid, 1, PCol));
    if StockBrush then
      Brush := SelectObject(ADC, GetStockObject(Null_Brush))
    else
      Brush := SelectObject(ADC, CreateSolidBrush(BCol));
    Reseted := False;
  end;
end;

procedure ResetColor(SaveColor: ColorData);
begin
  with SaveColor do
  begin
    if not Reseted then
    begin
      if StockPen then
        SelectObject(DC, Pen)
      else
        DeleteObject(SelectObject(DC, Pen));
      if StockBrush then
        SelectObject(DC, Brush)
      else
        DeleteObject(SelectObject(DC, Brush));
      Reseted := True;
    end;
  end;
end; 

function CMYK(c, m, y, k: Real): LongInt;
 var
    r, g, b: Real;
begin
  k := 1.0-k;
  r := k-c;
  g := k-m;
  b := k-y;
  r := int(255.5*r);
  g := int(255.5*g);
  b := int(255.5*b);
  r := min(255,max(0,r));
  g := min(255,max(0,g));
  b := min(255,max(0,b));
  CMYK := RGB(Trunc(r), Trunc(g), Trunc(b));
end;



var
  SaveExit: Pointer;

procedure GrToolsExit;
far;
begin
  ExitProc := SaveExit;
  if InfoPtr<>nil then
    Dispose(InfoPtr, Done);
end;

begin
  InfoPtr := nil;
  SaveExit := ExitProc;
  ExitProc := @GrToolsExit;
end.

