unit MahJongg;

interface

uses Objects, GrTools, MjDlgs, WinTypes, WinProcs, mmSystem, mjDefs, Strings, OWindows;

const
{STARTRAND =
     Zufall                       : 0
     c1 paarig                    : 34
     c2 paarig                    : 33
     c3 paarig                    : 5, 17, 18
}
    StartRand = 0;

    cl_Nothing  = 0;
    cl_Tschi    = 1;
    cl_Pong     = 2;
    cl_Kan      = 3;
    cl_MahJongg = 4;

    cl_Paar     = 5;
    cl_PaarigesSpiel = 6;

    tk_Spieler     = -1;
    tk_noTaktik    =  0;  {Pong/Kan Taktik}
    tk_Farbenspiel =  1;  {Katrin-Taktik}
    tk_Paarig      =  2;  {Parchen sammeln}

    sf_Unbekannt   = 0;
    sf_Teller      = 1;
    sf_Schrift     = 2;
    sf_Bambus      = 3;
    sf_Alle        = 4;

    NoTile       = 144;
    HalbVerdeckt = 145;
    Verdeckt     = 146;
type
    Tile = 0..146;
    Str2 = String[2];

    PTileArray = ^TTileArrayType;
    TTileArrayType = array[0..143] of tile;

    PPlayer = ^TPlayerClass;
    PComputerPlayer = ^TComputerPlayerClass;
    POpen = ^TOpenClass;

    TTileCollectionClass = Object(TObject)
      MaxTiles,
      Count: Integer;
      Tiles: PTileArray;
      constructor Init(MaxT: Integer);
      procedure   AddTile(T: Tile); virtual;
      procedure   InsertBefore(n: Integer; T: Tile); virtual;
      procedure   SortTiles; virtual;
      function    GetTile(n: Integer): Tile; virtual;
      function    GetIndex(S: Str2): Integer; virtual;
      function    CountTile(S: Str2): Integer; virtual;
      procedure   RemoveTile(n: Integer); virtual;
      procedure   Reset; virtual;
      destructor  Done; virtual;
    end;

    TWallClass = object(TObject)
      Tiles: TTileArrayType;
      Living, Dead: Integer;
      constructor Init;
      procedure   MergeTiles; virtual;
      function    DrawLivingTile(Who: Integer): Tile; virtual;
      function    DrawDeadTile(Which: Integer): Tile; virtual;
      procedure   Reset; virtual;
    end;

    TOpenClass = object(TObject)
      Tiles: TTileArrayType;
      Count: Integer;
      constructor Init;
      procedure   AddTile(T: Tile); virtual;
      function    GetTile(n: Integer): Tile; virtual;
      procedure   RemoveTile(n: Integer); virtual;
      function    CountOpen(TName: str2): Integer; virtual;
      procedure   Reset; virtual;
    end;

    TPlayerClass = Object(TObject)
      Name: String[20];
      Hand: TTileCollectionClass; {0..13}
      Open: TTileCollectionClass; {0..31}
      Punkte, GesamtPunkte, BestPunkte: LongInt;
      NextThrow: Integer;
      ThisCall: Integer;
      Changed: Boolean;
      Nummer: Integer;
      Taktik: Integer;
      Sammelfarbe: Integer;
      constructor Init(AName: String; Nr: Integer);
      procedure   DrawTile(T: Tile); virtual;
      procedure   DrawBonusTile(T: Tile); virtual;
      function    ThrowTile: Tile; virtual;
      function    Call(T: Tile): Integer; virtual;
      function    CalledTile(T: Tile; CL: Integer): Boolean; virtual;
      function    DoMahJongg(ST: Tile): Boolean;
      procedure   OpenHand; virtual;
      procedure   SelectTaktik; virtual;
      procedure   CalcPoints; virtual;
      procedure   Reset; virtual;
      destructor  Done; virtual;
    end;

    TComputerPlayerClass = Object(TPlayerClass)
      TschiStart: Integer;
      constructor Init(AName: String; Nr: Integer);
      function    ThrowTile: Tile; virtual;
      function    Call(T: Tile): Integer; virtual;
      function    CalledTile(T: Tile; CL: Integer): Boolean; virtual;
      procedure   SelectTaktik; virtual;
      procedure   Reset; virtual;
    end;


    TGameClass = Object(TObject)
      Wall: TWallClass;
      Open: TOpenClass;
      Player: array[0..3] of PPlayer;
      East, EastCount, CurrentPlayer, StartPos: Integer;
      Finished, WaitForThrow, WaitForCall, WaitForWeiter: Boolean;
      Austeilen, Austeilen2: Integer;
      Winner, CalcPt: Integer;
      AnzSpiele: Integer;
      Save:  record
               Player: array[0..3] of PPlayer;
               Living, Dead, OpenCount: Integer;
               Valid: Boolean;
               RSeed: LongInt;
             end;  
      constructor Init;
      procedure   NextGame; virtual;
      procedure   DoMove; virtual;
      function    GetWallHeight(Nr, Pos: integer): Integer; virtual;
      function    CountOpenTile(TName: Str2): Integer; virtual;
      procedure   ShowHighscore; virtual;
      procedure   SetHighscore; virtual;
      procedure   SaveGame; virtual;
      procedure   LoadGame; virtual;
      procedure   Reset; virtual;
      destructor  Done; virtual; 
    end;

var TheGame: TGameClass;                            


const
    TileName: array[Tile] of String[2] = (
       'X1', 'X2', 'X3', 'X4', 'Y1', 'Y2', 'Y3', 'Y4',
       'E',  'E',  'E',  'E',  'S',  'S',  'S',  'S',
       'W',  'W',  'W',  'W',  'N',  'N',  'N',  'N',
       'P',  'P',  'P',  'P',  'C',  'C',  'C',  'C',
       'F',  'F',  'F',  'F',  'T1', 'T1', 'T1', 'T1',
       'T2', 'T2', 'T2', 'T2', 'T3', 'T3', 'T3', 'T3',
       'T4', 'T4', 'T4', 'T4', 'T5', 'T5', 'T5', 'T5',
       'T6', 'T6', 'T6', 'T6', 'T7', 'T7', 'T7', 'T7',
       'T8', 'T8', 'T8', 'T8', 'T9', 'T9', 'T9', 'T9',
       'S1', 'S1', 'S1', 'S1', 'S2', 'S2', 'S2', 'S2',
       'S3', 'S3', 'S3', 'S3', 'S4', 'S4', 'S4', 'S4',
       'S5', 'S5', 'S5', 'S5', 'S6', 'S6', 'S6', 'S6',
       'S7', 'S7', 'S7', 'S7', 'S8', 'S8', 'S8', 'S8',
       'S9', 'S9', 'S9', 'S9', 'B1', 'B1', 'B1', 'B1',
       'B2', 'B2', 'B2', 'B2', 'B3', 'B3', 'B3', 'B3',
       'B4', 'B4', 'B4', 'B4', 'B5', 'B5', 'B5', 'B5',
       'B6', 'B6', 'B6', 'B6', 'B7', 'B7', 'B7', 'B7',
       'B8', 'B8', 'B8', 'B8', 'B9', 'B9', 'B9', 'B9',
       '-',  'HV', 'V');




implementation 



{------------ HILFSFUNKTIONEN --------------------------}


procedure Beep(S: String);
begin
  if Ton>0 then
  begin
    if (Ton=1) or (S='') then
      MessageBeep(0)
    else
    begin
      if S='MAHJONGG' then
        sndPlaySound('tataa.wav', snd_Sync)
      else
      if S='UNENTSCHIEDEN' then
        sndPlaySound('akkord.wav', snd_Sync)
      else
      if S='SPIELERZUG' then
        sndPlaySound('ding.wav', snd_Sync)
      else
        MessageBeep(0);
    end;
  end;
end;


function formatN(n: LongInt; Stellen: Integer): String;
 var
    s: String;
    i: Integer;
begin
  s := NToS(n);
  i := 2*(Stellen-length(s));
  if n<0 then
    i := i+1;
  s := copy('                  ', 1, i)+s;
  formatN := s;
end;

function GetFarbNr(T: Tile): Integer;
begin
  GetFarbNr := T div 36;
end;


{------------ TTILECOLLECTIONCLASS --------------------------}

constructor TTileCollectionClass.Init(MaxT: Integer);
 var
    i: Integer;
begin
  MaxTiles := MaxT;
  getMem(Tiles, MaxTiles*(SizeOf(TTileArrayType) div 144));
  for i := 0 to MaxTiles-1 do
    Tiles^[i] := NoTile;
  Count := 0; 
end;

procedure TTileCollectionClass.AddTile(T: Tile);
begin
  if Count<MaxTiles then
  begin
    Tiles^[Count] := T;
    Count := Count+1;
  end;
end;

procedure TTileCollectionClass.InsertBefore(n: Integer; T: Tile);
 var
    i: Integer;
begin
  if (Count<MaxTiles) and (n>=0) and (n<=Count) then
  begin
    for i := Count downto n+1 do
      Tiles^[i] := Tiles^[i-1];
    Tiles^[n] := T;
    Count := Count+1
  end;
end;

procedure TTileCollectionClass.SortTiles;
 var i, j: Integer;
     HelpTile: Tile;
begin
  for i := Count-1 downto 1 do         { simple as Bubblesort,}
    for j := 1 to i do                 {   but also sufficient}
      if Tiles^[j-1]>Tiles^[j] then
      begin
        HelpTile := Tiles^[j];
        Tiles^[j] := Tiles^[j-1];
        Tiles^[j-1] := HelpTile;
      end;
end;

function TTileCollectionClass.GetTile(n: Integer): Tile;
begin
  GetTile := Tiles^[n];
end;

function TTileCollectionClass.GetIndex(S: Str2): Integer; 
 var
    i: Integer;
begin
  GetIndex := -1;
  for i := 0 to Count-1 do
    if TileName[Tiles^[i]]=S then
      GetIndex := i;
end;

function TTileCollectionClass.CountTile(S: Str2): Integer;
 var
    i, Anz: Integer;
begin
  Anz := 0;
  for i := 0 to Count-1 do
    if TileName[Tiles^[i]]=S then
      Anz := Anz+1;
  CountTile := Anz;
end;


procedure TTileCollectionClass.RemoveTile(n: Integer);
 var
    i: Integer;
begin
  if n<Count then
  begin
    for i:= n to Count-2 do
      Tiles^[i] := Tiles^[i+1];
    Tiles^[Count-1] := NoTile;
    Count := Count-1;
  end;
end;

procedure TTileCollectionClass.Reset;
 var
    i: Integer;
begin
  for i := 0 to Count-1 do
    Tiles^[i] := NoTile;
  Count := 0;
end;

destructor TTileCollectionClass.Done;
begin
  FreeMem(Tiles,MaxTiles*(SizeOf(TTileArrayType) div 144));
end;



{------------ TWALLCLASS --------------------------}

constructor TWallClass.Init;
 var
    i: Integer;
begin
  Reset;
end;

procedure TWallClass.MergeTiles;
 var
    i, n: Integer;
    HelpTile: Tile; 
begin
  for i := 143 downto 1 do
  begin
    n := random(i+1);
    HelpTile := Tiles[i];
    Tiles[i] := Tiles[n];
    Tiles[n] := HelpTile;
  end;
end;

function TWallClass.DrawLivingTile(Who: Integer): Tile;
 var
    i: Integer;
    t: Tile;
begin
  if (Who=0) and (TheGame.Player[0]^.Name='Mogel') then
  begin
    i := Living;
    while i<>-1 do
    begin
      T := Tiles[i];
      if AppMessageString('Mchten Sie Stein '+NToS(i)+': '+TileName[T],'MOGEL', mb_YesNo+mb_IconQuestion)=id_Yes then
      begin
        Tiles[i] := Tiles[Living];
        Tiles[Living] := T;
        i := -1;
      end
      else
        i := i+1;
      if i=Dead then
      begin
        AppMessage('Nochmal von vorne:');
        i := Living;
      end;
    end;
  end;
  DrawLivingTile := Tiles[Living];
  Living := Living+1;
end; 

function TWallClass.DrawDeadTile(Which: Integer): Tile;
begin
  DrawDeadTile := Tiles[Dead];
  Dead := Dead-1;
end;
 
procedure TWallClass.Reset;
 var
    i: integer;
begin
  for i := 0 to 143 do
    Tiles[i] := i;
  Living := 0;
  Dead := 143;
  MergeTiles;
end;



{------------ TOPENCLASS --------------------------}

constructor TOpenClass.Init;
begin
  Reset;
end;

procedure TOpenClass.AddTile(T: Tile);
begin
  Tiles[Count] := T;
  Count := Count+1;
end;
 
function TOpenClass.GetTile(n: Integer): Tile;
begin
  if (n>=0) and (n<Count) then
    GetTile := Tiles[n]
  else
    GetTile := NoTile;
end;

procedure TOpenClass.RemoveTile(n: Integer);
 var
    i: Integer;
begin
  if (n>=0) and (n<Count) then
  begin
    for i:= n to Count-2 do
      Tiles[i] := Tiles[i+1];
    Tiles[Count-1] := NoTile;
    Count := Count-1;
  end;
end;


function TOpenClass.CountOpen(TName: Str2): Integer;
 var
    i, n: Integer;
begin
  n := 0;
  for i := 0 to Count-1 do
    if TileName[Tiles[i]]=TName then
      n := n+1;
  CountOpen := n; 
end;
 
procedure TOpenClass.Reset; 
 var
    i: Integer;
begin
  for i := 0 to 143 do
    Tiles[i] := NoTile;
  Count := 0;
end;


{------------ TPLAYERCLASS --------------------------}

constructor TPlayerClass.Init(AName: String; Nr: integer);
begin
  Name := AName;
  Nummer := Nr;
  Hand.Init(14);
  Open.Init(37);
  Punkte := 0;
  GesamtPunkte := 0;
  NextThrow := 0;
  Changed := True;
  Taktik := tk_Spieler;
end;

procedure TPlayerClass.DrawBonusTile(T: Tile);
 var
    i: Integer;
begin
  Changed := True;
  Open.InsertBefore(0, T);
end;

procedure TPlayerClass.DrawTile(T: Tile);
begin
  Changed := True;
  Hand.AddTile(T);
end;


function TPlayerClass.ThrowTile: Tile;
begin
  if NextThrow=-1 then
    ThrowTile := NoTile
  else
  begin
    Changed := True;
    ThrowTile := Hand.GetTile(NextThrow);
    Hand.RemoveTile(NextThrow);
    Hand.SortTiles;
  end;
end;


function TPlayerClass.Call(T: Tile): Integer;
begin
  Call := ThisCall;
end;

procedure TPlayerClass.Reset;
begin
  Changed := True;
  Hand.Reset;
  Open.Reset;
  Punkte := 0;
  NextThrow := 0;
end;


function TPlayerClass.CalledTile(T: Tile; CL: Integer): Boolean;
 var
    i, j, k, Anz, Nr, first, last, Start: integer;
    IsOK, Found: Boolean;
    TName: String[2];
    Farbe: Char;
    CalledOK: Boolean;
    F: Integer;
begin
  CalledOK := False;

  F := sf_Unbekannt;
    
  if T=NoTile then
  begin
    case CL of
      cl_Kan:
        begin
          T := Hand.GetTile(Hand.Count-1);
          TName := TileName[T];
          if Open.CountTile(TName)=3 then
          begin
            F := GetFarbNr(T);
            CalledOK := True;
            i := Open.GetIndex(TName);
            Open.InsertBefore(i-1, HalbVerdeckt);
            Open.InsertBefore(i-1, T);
            Open.InsertBefore(i-1, HalbVerdeckt);
            Hand.RemoveTile(Hand.Count-1);
          end
          else
          begin
            i := 0;
            while (i<Hand.Count) do
            begin
              T := Hand.GetTile(i);
              TName := TileName[T];
              if Hand.CountTile(TName)=4 then
              begin
                F := GetFarbNr(T);
                CalledOK := True;
                Open.AddTile(NoTile);
                for k := 1 to 4 do
                begin
                  j := Hand.GetIndex(TName);
                  if k=4 then
                    Open.AddTile(Verdeckt);
                  Open.AddTile(Hand.GetTile(j));
                  Hand.RemoveTile(j);
                end;
                i := Hand.Count
              end;
              i := i+1;
            end;
          end;
          if CalledOK and (Nummer<>0) then
          begin
            Beep('');
            InfoTxt := name+' hat einen Kan';
          end;
        end;
      cl_MahJongg: CalledOK := DoMahJongg(NoTile);
    end;
  end
  else
  begin
    TName := TileName[T];
    case CL of
      cl_Tschi: begin
                  Farbe := TName[1];
                  if (Farbe in ['B', 'S', 'T']) and (length(TName)=2) then
                  begin
                    Nr := ord(Tname[2])-ord('0');
                    first := IMax(1,Nr-2);
                    last := IMin(7,Nr);
                    Found := False;
                    Start := 0;
                    for i := first to last do
                    begin
                      if not Found then
                      begin
                        IsOK := True;
                        for j := i to i+2 do
                          if (j<>Nr) and (Hand.CountTile(Farbe+chr(j+ord('0')))=0) then
                            IsOK := False;
                        if IsOK then
                        begin
                          if Start<>0 then
                          begin
                            if AppMessageString('Tschi von '+NToS(Start)+' bis '+NToS(Start+2)+'?','TSCHI',
                                                mb_IconQuestion+mb_YesNo)=id_Yes then
                              Found := True
                            else
                              Start := i;
                          end
                          else
                            Start := i;
                        end;
                      end;
                    end;
                    if Start>0 then
                    begin
                      F := GetFarbNr(T);
                      CalledOK := True;
                      Open.AddTile(NoTile);
                      for i := Start to Start+2 do
                      begin
                        if i=Nr then
                          Open.AddTile(T)
                        else
                        begin
                          j := Hand.GetIndex(Farbe+chr(i+ord('0')));
                          Open.AddTile(Hand.GetTile(j));
                          Hand.RemoveTile(j);
                        end;
                      end;
                    end;
                  end;
                end;
      cl_Pong: begin
                 Anz := 0;
                 for i := 0 to Hand.Count-1 do
                 begin
                   if TileName[Hand.GetTile(i)]=Tname then
                     Anz := Anz+1;
                 end;
                 if Anz>=2 then
                 begin
                   CalledOK := True;
                   F := GetFarbNr(T);
                   Open.AddTile(NoTile);
                   Open.AddTile(T);
                   Anz := 2;
                   i := 0;
                   While (Anz>0) and (i<Hand.Count) do
                   begin
                     if TileName[Hand.GetTile(i)]=Tname then
                     begin
                       Open.AddTile(Hand.GetTile(i));
                       Hand.RemoveTile(i);
                       Anz := Anz-1;
                     end
                     else
                       i := i+1;
                   end;
                 end;
               end;
      cl_Kan: begin
                 Anz := 0;
                 for i := 0 to Hand.Count-1 do
                 begin
                   if TileName[Hand.GetTile(i)]=Tname then
                     Anz := Anz+1;
                 end;
                 if Anz>=3 then
                 begin
                   CalledOK := True;
                   F := GetFarbNr(T);
                   Open.AddTile(NoTile);
                   Open.AddTile(T);
                   Anz := 3;
                   i := 0;
                   While (Anz>0) and (i<Hand.Count) do
                   begin
                     if TileName[Hand.GetTile(i)]=Tname then
                     begin
                       Open.AddTile(Hand.GetTile(i));
                       Hand.RemoveTile(i);
                       Anz := Anz-1;
                     end
                     else
                       i := i+1;
                   end;
                 end;
               end;
      cl_MahJongg: CalledOK := DoMahJongg(T);
    end;
  end;
  CalledTile := CalledOK;
  if CalledOK then
  begin
    if F <> sf_Unbekannt then
    begin
      if SammelFarbe = sf_Unbekannt then
        SammelFarbe := F
      else
      if SammelFarbe <> F then
        SammelFarbe := sf_alle;
    end;
    Changed := True;
  end;
end;

function TPlayerClass.DoMahJongg(ST: Tile): Boolean;
 var
    SchlussPaar, Paarig: Boolean;
    n: Integer;

  function RekMahJongg: Boolean;
   var
      i, j, Anz: integer;
      Temp: TTileCollectionClass;
      T: Tile;
      TName: Str2;

  procedure Temp2Open;
   var
      i: Integer;
  begin
    Open.AddTile(NoTile);
    for i := 0 to Temp.Count-1 do
    begin
      if i=Temp.Count-1 then
        Open.AddTile(Verdeckt);
      Open.AddTile(Temp.GetTile(i));
    end;
    Temp.Done;
  end;

  begin
    if Hand.Count=0 then
      RekMahJongg := True
    else
    begin
      Temp.Init(3);
      T := Hand.GetTile(0);
      TName := TileName[T];
      Hand.RemoveTile(0);
      Temp.AddTile(T);
      Anz := Hand.CountTile(TileName[T]);

      {PONG}
      if Anz>=2 then 
      begin
        i := Hand.GetIndex(TName);
        Temp.AddTile(Hand.GetTile(i));
        Hand.RemoveTile(i);
        i := Hand.GetIndex(TName);
        Temp.AddTile(Hand.GetTile(i));
        Hand.RemoveTile(i);
        if RekMahJongg then
        begin
          Temp2Open;
          RekMahJongg := true;
          exit;
        end
        else
        begin
          Hand.InsertBefore(0,Temp.GetTile(2));
          Temp.RemoveTile(2);
          Hand.InsertBefore(0,Temp.GetTile(1));
          Temp.RemoveTile(1);
        end;
      end;

      {SchlussPaar}
      if (Anz>=1) and not Schlusspaar then 
      begin
        SchlussPaar := True;
        i := Hand.GetIndex(TName);
        Temp.AddTile(Hand.GetTile(i));
        Hand.RemoveTile(i);
        if RekMahJongg then
        begin
          Temp2Open;
          RekMahJongg := true;
          exit;
        end
        else
        begin
          SchlussPaar := False;
          Hand.InsertBefore(0, Temp.GetTile(1));
          Temp.RemoveTile(1);
        end;
      end;

      {TSCHI}
      if (TName[1] in ['B', 'S', 'T']) and (length(TName)=2) and (TName[2]<='7') then
      begin
        i := Hand.GetIndex(TName[1]+Chr(Ord(TName[2])+1));
        if i<>-1 then
        begin
          j := Hand.GetIndex(TName[1]+Chr(Ord(TName[2])+2));
          if j<>-1 then
          begin
            Temp.AddTile(Hand.GetTile(i));
            Temp.AddTile(Hand.GetTile(j));
            if i<j then
            begin
              Hand.RemoveTile(j);
              Hand.RemoveTile(i);
            end
            else
            begin
              Hand.RemoveTile(i);
              Hand.RemoveTile(j);
            end;
            if RekMahJongg then
            begin
              Temp2Open;
              RekMahJongg := True;
              exit
            end;
          end;
        end; 
      end;
      for i := 0 to Temp.Count-1 do
        Hand.AddTile(Temp.GetTile(i));
      Hand.SortTiles;
      Temp.Done;
      RekMahJongg := False;
    end;
  end;

begin
  if ST<>NoTile then
    Hand.AddTile(ST);
  Hand.SortTiles;

  {Test auf Paariges Spiel}
  if Hand.Count=14 then
  begin
    Paarig := True;
    for n := 0 to 13 do
      if odd(Hand.CountTile(TileName[Hand.GetTile(n)])) then
        Paarig := False;
    if Paarig then
    begin
      OpenHand;
      Changed := True;
      DoMahJongg := True;
      exit;
    end;
  end;

  SchlussPaar := False;
  if RekMahJongg then
  begin
    DoMahJongg := True;
    Changed := True;
  end
  else
    DoMahJongg := False;
end;


procedure TPlayerClass.OpenHand;
 var
    Anz, i, j: Integer;
    T: Tile;
    Paarig: Boolean;
begin
  if Hand.Count=14 then   {Test auf Paariges Spiel}
  begin
    Paarig := True;
    for i := 0 to 13 do
      if odd(Hand.CountTile(TileName[Hand.GetTile(i)])) then
        Paarig := False;
    if Paarig then
    begin
      Open.AddTile(NoTile);
      for i := 0 to 13 do
        Open.AddTile(Hand.GetTile(i));
      for i := 0 to 13 do
        Hand.RemoveTile(0);
    end;
  end;
  while Hand.Count>0 do
  begin
    T := Hand.GetTile(0);
    Anz := Hand.CountTile(TileName[T]);

    {Pong auf der Hand rauslegen}
    if Anz>=3 then
    begin
      Open.AddTile(NoTile);
      j:= Hand.GetIndex(TileName[T]);
      Open.AddTile(Hand.GetTile(j));
      Hand.RemoveTile(j);

      j:= Hand.GetIndex(TileName[T]);
      Open.AddTile(Hand.GetTile(j));
      Hand.RemoveTile(j);

      Open.AddTile(Verdeckt);
      j:= Hand.GetIndex(TileName[T]);
      Open.AddTile(Hand.GetTile(j));
      Hand.RemoveTile(j);
    end
    else
    {Trumpfpaare rauslegen}
    if (Anz=2) and (T<36) then
    begin
      Open.AddTile(NoTile);
      j:= Hand.GetIndex(TileName[T]);
      Open.AddTile(Hand.GetTile(j));
      Hand.RemoveTile(j);

      Open.AddTile(Verdeckt);
      j:= Hand.GetIndex(TileName[T]);
      Open.AddTile(Hand.GetTile(j));
      Hand.RemoveTile(j);
    end
    else
      Hand.RemoveTile(0);
  end;
  Changed := True;
end;


procedure TPlayerClass.SelectTaktik;
begin
  Taktik := tk_Spieler;
end;



Procedure TestFarbe(var F: Char; T: Tile);
 var
    F2: String;
begin
  if T<36 then
    exit;
  if F='-' then
    exit;
  F2:=copy(TileName[T],1,1);
  if F=F2 then
    exit;
  if F='?' then
    F := F2[1]
  else
    F := '-';
end;

procedure TPlayerClass.CalcPoints;
 var
    s2, s3: String;
    a: PChar;
    T, FirstT, Mode: Tile;
    Anz, i, j, Art: integer;
    P, V, SumP, SumV: Integer;

    AnzTschi, AnzPong, AnzTrumpf, AnzX, AnzY: Integer;
    TrumpfPaar: Boolean;
    Sonder: Char;
    Farbe: Char;

  procedure AddLine(Punkte: LongInt; Verdopplungen: Integer; Txt: String);
   var
      s: String;
  begin
    s := FormatN(Punkte,5)+' P '+FormatN(Verdopplungen,2)+' V  - '+Txt+#13#10;
    StrPCat(a, s);
    SumP := SumP+Punkte;
    SumV := SumV+Verdopplungen;
  end;

begin
  GetMem(a, 4096);

  Sonder := '-';
  AnzX := 0;
  AnzY := 0;
  AnzTschi := 0;
  AnzPong := 0;
  AnzTrumpf := 0;
  TrumpfPaar := False;
  Farbe := '?';

  StrPCopy(a,'Punkte fr '+name+#13#10#13#10);
  SumP := 0;
  SumV := 0;
  P := 0;
  V := 0;
  i := 0;

  {Blumensteine Zhlen 4 Punkte}

  while (i<Open.Count) and (Open.GetTile(i)<8) do
  begin
    P := P+4;
    s3 := TileName[Open.GetTile(i)];
    if s3[1]='X' then
      AnzX := AnzX+1
    else
      AnzY := AnzY+1;
    if ord(s3[2])-ord('1')=(Nummer+4-TheGame.East) mod 4 then
      V := V+1;
    i := i+1;
  end;
  if P>0 then
    AddLine(P, V, 'Blumensteine');

  if (AnzX = 4) and (AnzY=4) then
    AddLine(0, 3, 'Blumenstrau')
  else
  if (AnzX=4) or (AnzY=4) then
    AddLine(0, 1, 'Blumenstein Kan');

  Art := cl_Nothing;
  while i<=Open.Count do
  begin
    if i=Open.Count then
      T := NoTile
    else
      T := Open.GetTile(i);
    if T=NoTile then
    begin
      P := 0;
      V := 0;
      if Art<>cl_Nothing then
      begin
        case Art of
          cl_Paar: begin
                     s2 := '';
                     if FirstT<36 then
                     begin
                       s2 := 'Trumpf-';
                       P := 2;
                       TrumpfPaar := True;
                     end;
                     Mode := NoTile;
                     s2 := s2+'Parchen von '+TileName[FirstT];
                     TestFarbe(Farbe, FirstT);
                   end;
          cl_Tschi:begin
                     s2 := 'Tschi ab '+TileName[FirstT];
                     TestFarbe(Farbe, FirstT);
                     AnzTschi := AnzTschi+1;
                   end;
          cl_Pong: begin
                     P := 2;
                     s2 := 'Pong von '+TileName[FirstT];
                     TestFarbe(Farbe, FirstT);
                     AnzPong := AnzPong+1;
                   end;
          cl_Kan:  begin
                     P := 8;
                     V := 1;
                     s2 := 'Kan von '+TileName[FirstT];
                     TestFarbe(Farbe, FirstT);
                   end;
          cl_PaarigesSpiel:
                   begin
                     P := 2000;
                     s2 := 'Paariges Spiel';
                     Farbe := '-';
                     Sonder := 'P';
                   end;
        end;
        if (Art=cl_Pong) or (Art=cl_Kan) then
        begin
          if FirstT<36 then
          begin
            s2 := 'Trumpf-'+s2;
            P := P*2;
            V := V+1;
            if pos(TileName[FirstT],'ESWN')-1 = ((4-TheGame.East+Nummer) mod 4) then
            begin
              s2 := 'eigener '+s2;
              V := V+1;
            end;
            AnzTrumpf := AnzTrumpf+1;
          end;
          if Mode=Verdeckt then
          begin
            s2 := 'verdeckter '+s2;
            P := P*2;
          end;
          if Mode=Halbverdeckt then
          begin
            s2 := 'halbverdeckter '+s2;
            P := P*2;
          end;
          if (length(TileName[FirstT])=2) and
               ((copy(TileName[FirstT],2,1)='1')  or
                (copy(TileName[FirstT],2,1)='9')) then
          begin
{            s2 := s2+' (von 1 oder 9)';}
            P := P*2;
          end;
        end;
        AddLine(P, V, s2);
      end;

      Anz := 0;
      Mode := NoTile;
      Art := cl_nothing;
      FirstT := NoTile;
    end
    else
    begin
      if T=Verdeckt then
        Mode := Verdeckt
      else
      if T=HalbVerdeckt then
        Mode := Halbverdeckt
      else
      begin
        Anz := Anz+1;
        case Anz of
          1 : FirstT := T;
          2 : if TileName[T]<>TileName[FirstT] then
                Art := cl_Tschi
              else
                Art := cl_Paar;  
          3 : if Art<>cl_Tschi then
                Art := cl_Pong;
          4 : Art := cl_Kan;
         14 : Art := cl_PaarigesSpiel;
        end;  
      end;
    end;
    i := i+1;
  end;

  if TheGame.Winner = Nummer then
  begin
    if Sonder='-' then
      AddLine(20, 0, 'Gewonnen');
    if AnzTschi=4 then
      AddLine(2, 0, 'nur Tschi');
    if AnzPong=4 then
      AddLine(10, 0, 'nur Pong');
    if (Farbe<>'-') and (AnzTrumpf<=2) then
    begin
      if (AnzTrumpf=0) and not TrumpfPaar then
        AddLine(0, 4, 'reines Farbenspiel')
      else
      if (AnzTrumpf=0) and TrumpfPaar then
        AddLine(0, 3, 'reines Farbenspiel mit Trumpf-Paar')
      else
      if AnzTrumpf=1 then
        AddLine(0, 2, 'unreines Farbenspiel (1 Trumpf)')
      else
      if AnzTrumpf=2 then
        AddLine(0, 1, 'unreines Farbenspiel (2 Trumpf)');
    end;
  end;

  Punkte := SumP;
  for i := 1 to SumV do
    Punkte := Punkte*2;
  StrPCat(a, '__________'+#13#10+
             FormatN(SumP,5)+' P '+FormatN(SumV,2)+' V  = '+NToS(Punkte)+' Punkte');
  MessageBox(Application^.Mainwindow^.HWindow, a, 'Punkte', mb_OK);
  FreeMem(a, 4096);
end;

destructor TPlayerClass.Done;
begin
  Hand.Done;
  Open.Done;
end;

{------------ TCOMPUTERPLAYERCLASS --------------------------}


constructor TComputerPlayerClass.Init(AName: String; Nr: Integer);
begin
  TPlayerClass.Init(AName, Nr);
end;

function TComputerPlayerClass.ThrowTile: Tile;
 var
    T: Tile;
    i, j, k, AnzHand, AnzOpen, AH, AO: integer;
    p: array[0..13] of integer;
    FAnz: array[0..3,0..4] of Integer;
    MinWert, MinJ, Found: Integer;
begin
  Hand.SortTiles;

  { Blumensteine sofort rauslegen (egal welche Taktik) }
  if Hand.GetTile(0)<8 then     
    k := 0
  else
  begin

    { Zufallspermutation berechnen, kann spter in jeder Taktik
    benutzt werden. }
    for i := 0 to Hand.Count-1 do
      p[i] := i;
    for i := Hand.Count downto 2 do
    begin
      j := random(i);
      k := p[i-1];
      p[i-1] := p[j];
      p[j] := k;
    end;

    { Taktik zum Werfen anwenden: }
    case Taktik of

      { Die Pong/Kan - Taktik }
      tk_NoTaktik :  
          Begin
            k := Hand.Count-1;
            AnzHand := 4;
            AnzOpen := 0;
            for i := 0 to Hand.Count-1 do
            begin
              j := p[i];
              AH := Hand.CountTile(TileName[Hand.GetTile(j)]);
              AO := TheGame.CountOpenTile(TileName[Hand.GetTile(j)]);
              if AO=1 then
                AO := 0;
              if AH<AnzHand then
              begin
                k := j;
                AnzHand := AH;
                AnzOpen := AO;
              end
              else
              if (AH=AnzHand) and (AO>AnzOpen) then
              begin
                k := j;
                AnzOpen := AO;
              end;
            end;
          end;

      { Die Katrin-Taktik }
      tk_Farbenspiel :
          begin
            if SammelFarbe = sf_Alle then
            begin                        { Alle Farben drfen geschmissen werden }
              for i := 1 to 3 do
                FAnz[i,0] := 0;
            end
            else if (SammelFarbe <> sf_Unbekannt) then 
            begin                        { nur Farbe SFarbe darf nicht geschmissen werden }
              for i := 1 to 3 do
                FAnz[i,0] := 0;
              FAnz[SammelFarbe,0] := 1;  
            end
            else  {SammelFarbe = sf_Unbekannt}
            begin
              { schlechteste Farbe ermitteln und andere Farben zu
                Sammelfarbe erklren. }

              { Zhlen der Farbigen Steine (einzeln, Paar,
                                            Pong und Kan) } 
              for i := 0 to 3 do           { FAnz zurcksetzen}
                for j := 0 to 4 do
                  FAnz[i, j] := 0;

              for i := 0 to Hand.Count-1 do
              begin
                T := Hand.GetTile(i);
                j := Hand.CountTile(TileName[T]);
                k := GetFarbNr(T);
                FAnz[k, j] := FAnz[k, j]+1;
              end;

              Found := 0;
              for i := 1 to 3 do
              begin
                if FAnz[i, 3]+FAnz[i, 4]>0 then
                begin
                  FAnz[i,0] := 1;
                  Found := Found+1;
                end;  
              end;
              if Found=0 then     { kein Farbe mit Pong oder Kan }
              begin
                k := 999;         { berechne Bewertung fr alle Farben, k=Minimum (auer 0) }
                for i:= 1 to 3 do
                begin
                  FAnz[i,3] := 4*FAnz[i, 1]+7*FAnz[i,2];
                  if (FAnz[i,3]<k) and (FAnz[i,3]<>0) then
                    k := FAnz[i,3];
                end;
                for i := 1 to 3 do  { markiere Farben, die nicht minimale Bewertung haben }
                  if FAnz[i,3] <> k then
                    FAnz[i,0] := 1;
              end
              else
              if Found=1 then { nur eine Farbe mit Pong oder Kan sammeln }
              begin
                for i := 1 to 3 do
                  if FAnz[i,3]+FAnz[i,4]>0 then
                    FAnz[i,0] := 1; 
              end;
                  {Found>=2 ->  mehrere Farben mit Pong oder Kan => alles sammeln }
            end;

            { FAnz[FarbNr, 0] = 0    :  Farbe wird nicht gesammelt,
              FAnz[FarbNr, 0] = 1    :  Farbe wird gesammelt.}

            for i := 1 to 3 do
              if FAnz[i,0]=1 then
                FAnz[i,0] := 2;
            FAnz[0,0] := 1;       { Trmpfe wie Sammelfarbe behandeln, sofern eine existiert }

{Message('SF: '+NToS(SammelFarbe)+' T:'+NTos(FAnz[1,0])+
                                 ' S:'+NTos(FAnz[2,0])+
                                 ' B:'+NTos(FAnz[3,0]));}
            MinWert := 9999;
            MinJ := 0;
            for i := 0 to Hand.Count-1 do
            begin                     { Berechne Wertung fr jeden Stein }
              j := p[i];
              T := Hand.GetTile(j);
              AnzHand := Hand.CountTile(TileName[T]);
              AnzOpen := TheGame.CountOpenTile(TileName[T]);
              k := FAnz[GetFarbNr(T), 0];
              case k of
                0 : k := AnzHand*2-AnzOpen+1;
                1 : k := AnzHand*2-AnzOpen+4;
                2 : k := AnzHand*2-AnzOpen+5;
              end;  
{Message(TileName[T]+': '+NToS(k));}
              if k<MinWert then
              begin
                MinWert := k;
                MinJ := j;
              end;
            end;
            k := MinJ;
          end;
      tk_Paarig :
          begin
            MinWert := 9999;
            MinJ := 0;
            for i := 0 to Hand.Count-1 do
            begin
              j := p[i];
              T := Hand.GetTile(j);
              AnzHand := Hand.CountTile(TileName[T]);
              AnzOpen := TheGame.CountOpenTile(TileName[T]);
              if Odd(AnzHand) then
                k := -AnzOpen-AnzHand
              else
                k := 10;
              if k<MinWert then
              begin
                MinWert := k;
                MinJ := j;
              end;
            end;
            k := MinJ;
          end;
    end;
  end;
  T := Hand.GetTile(k);
  Hand.RemoveTile(k);
  InfoTxt := name+' wirft '+TileName[T];
  ThrowTile := T;
end;


function TComputerPlayerClass.Call(T: Tile): Integer;
 var
    i, j, Nr, First, Last: Integer;
    F: Integer;
    FOk, Found: Boolean;
    TName: Str2;
begin
  Call := cl_Nothing;
  if T=NoTile then
  begin
    T := Hand.GetTile(Hand.Count-1);
    if (Taktik<>tk_Paarig) and ((Hand.CountTile(TileName[T])=4) or (Open.CountTile(TileName[T])=3)) then
      Call := cl_Kan
    else
    begin
      case Taktik of
        tk_NoTaktik:
            Call := cl_MahJongg;
        tk_FarbenSpiel:
            begin
              if (SammelFarbe=sf_alle) or (SammelFarbe=sf_Unbekannt) then
                Call := cl_MahJongg
              else
              begin
                FOk := True;
                for i := 0 to Hand.Count-1 do
                  if FOk then
                  begin
                    F := GetFarbNr(Hand.GetTile(i));
                    if F<>sf_Unbekannt then
                      FOK := (F=SammelFarbe);
                  end;
                if FOK then
                  Call := cl_MahJongg;
              end;
            end;
        tk_Paarig:
            Call := cl_MahJongg;
      end;
    end;
  end
  else
  begin
    if (Hand.CountTile(TileName[T])=1) then
    begin
      if (Hand.Count=1) then
      begin
        F := GetFarbNr(T);
        if (Taktik<>tk_Farbenspiel) or
             (F=sf_Unbekannt) or (SammelFarbe=sf_Alle) or
             (SammelFarbe=sf_Unbekannt) or (F=SammelFarbe) then
        begin
          Beep('');
          Call := cl_MahJongg;
          InfoTxt := name+' ruft MAH JONGG';
          exit;
        end;
      end
      else if (Taktik=tk_Paarig) and (Hand.Count=13) then
      begin
        FOk := True;
        for i := 0 to Hand.Count-1 do
          if (TileName[Hand.GetTile(i)]<>TileName[T]) and
             Odd(Hand.CountTile(TileName[Hand.GetTile(i)])) then
            FOk := False;
        if FOk then
        begin
          Beep('');
          Call := cl_MahJongg;
          InfoTxt := name+' ruft MAH JONGG';
          exit;
        end;
      end;
    end;
    if Taktik<>tk_Paarig then
    begin
      if Hand.CountTile(TileName[T])=2 then
      begin
        F := GetFarbNr(T);
        if (F=sf_Unbekannt) or (SammelFarbe=sf_Alle) or
           (SammelFarbe=sf_Unbekannt) or (F=SammelFarbe) then
        begin
          Beep('');
          Call := cl_Pong;
          InfoTxt := name+' ruft PONG';
          exit;
        end;
      end
      else
      if Hand.CountTile(TileName[T])=3 then
      begin
        Beep('');
        Call := cl_Kan;
        InfoTxt := name+' ruft KAN';
        exit;
      end
      else
      if (Hand.CountTile(TileName[T])=0) and ((TheGame.CurrentPlayer+1 mod 4)=Nummer) then
      begin
        F := GetFarbNr(T);
        if (F<>sf_Unbekannt) and
             ((Taktik<>tk_Farbenspiel)   or
              (SammelFarbe=sf_Alle)      or
              (F=SammelFarbe))              then
        begin
          Nr := ord(TileName[T][2])-ord('0');
          first := IMax(1,Nr-2);
          last := IMin(7,Nr);
          TschiStart := 0;
          for i := first to last do
          begin
            Found := True;
            for j := i to i+2 do
            begin
              TName := TileName[T][1]+chr(ord('0')+j);
              if (TName<>TileName[T]) then
              begin
                if (Hand.GetIndex(TName)=-1) then
                  found := False
                else
                  if (Hand.CountTile(TName)>1) and (TheGame.CountOpenTile(TName)<2) then
                    found := False;
              end;
            end;
            if Found then
            begin
              TschiStart := i;
              Beep('');
              Call := cl_Tschi;
              InfoTxt := name+' ruft Tschi';
              exit;
            end;
          end;
        end;
      end;
    end;
  end;
end;


function TComputerPlayerClass.CalledTile(T: Tile; CL: Integer): Boolean;
 var
    i, j: integer;
    TName: Str2;
begin
  if cl<>cl_Tschi then
    CalledTile := TPlayerClass.CalledTile(T, CL)
  else
  begin
    Open.AddTile(NoTile);
    for i := TschiStart to TschiStart+2 do
    begin
      TName :=  TileName[T][1]+chr(ord('0')+i);
      if TName = TileName[T] then
        Open.AddTile(T)
      else
      begin
        j := Hand.GetIndex(TName);
        Open.AddTile(Hand.GetTile(j));
        Hand.RemoveTile(j); 
      end;
    end;
    Changed := True;
    CalledTile := True;
  end
end;



procedure TComputerPlayerClass.SelectTaktik;
 var
    i, j, AnzPaar, AnzPong: integer;
    AnzFarbe: array[0..3] of Integer;
    T: Tile;
begin
  Taktik := tk_NoTaktik;
  sammelFarbe := sf_unbekannt;
  AnzPaar := 0;
  AnzPong := 0;
  for i := 0 to 3 do
    AnzFarbe[i] := 0;
  for i := 0 to Hand.Count-1 do
  begin
    T := Hand.GetTile(i);
    j := Hand.CountTile(TileName[T]);
    if j>=2 then
      AnzPaar := AnzPaar+1;
    if j=3 then
      AnzPong := AnzPong+1; 
    AnzFarbe[GetFarbNr(T)] := AnzFarbe[GetFarbNr(T)]+1;
  end;
  AnzPaar := (AnzPaar-(AnzPong div 3)) div 2;

  j := 0;
  for i := 1 to 3 do
    if AnzFarbe[i]>j then
      j := AnzFarbe[i];
  if j>=5 then
    Taktik := tk_Farbenspiel;
  if AnzPaar>=4 then
    taktik := tk_Paarig;
  if TheGame.Player[0]^.Name='Mogel' then
    AppMessage(name+': '+NToS(taktik));
end;



procedure TComputerPlayerClass.Reset;
begin
  TPlayerClass.Reset;
{  Taktik := tk_Farbenspiel;
  SammelFarbe := sf_unbekannt;}
end;


{------------ TGAMECLASS --------------------------}

constructor TGameClass.Init;
 var
    i: Integer;
begin
  if StartRand=0 then
    Randomize
  else
    RandSeed := StartRand;
  Wall.Init;
  Open.Init;
  Player[0] := New(PPlayer, Init('Spieler1', 0));
  for i := 1 to 3 do
    Player[i] := New(PComputerPlayer, Init('Computer'+NToS(i), i));

  Save.Player[0] := New(PPlayer, Init('SaveS1', 0));
  for i := 1 to 3 do
    Save.Player[i] := New(PComputerPlayer, Init('SaveC1'+NToS(i), i));
  Save.Valid := False;

  Reset;
end;

procedure TGameClass.NextGame;
 var
    i, j: Integer;
begin
  Wall.Reset;
  Open.Reset;
  for i := 0 to 3 do
    Player[i]^.Reset;

  AnzSpiele := AnzSpiele+1;

  Austeilen := 8;
  Austeilen2 := East;

  CurrentPlayer := East;
  Player[CurrentPlayer]^.NextThrow := -1;
  WaitForThrow := True;
  WaitForCall := False;
  Finished := False;

  Save.Valid := False;
end;

procedure TGameClass.DoMove;
 var
    i, j: Integer;
    T: Tile;
    s: String;
    MaxCall, PNr, CL: Integer;
    DoNotThrow: Boolean;
    Diff: LongInt;
    a: PChar;
begin
  if WaitForWeiter then
    exit;

  if Finished then
  begin
    if CalcPt>0 then
    begin
      for i := 0 to 3 do
        Player[i]^.CalcPoints;
      for i := 0 to 2 do
      begin
        for j := i+1 to 3 do
        begin
          Diff := Player[i]^.Punkte - Player[j]^.Punkte;
          if (i=East) or (j=East) then
            Diff := Diff*2;
          if ((i=Winner) and (Diff<0)) or
             ((j=Winner) and (Diff>0)) then
            Diff := 0;
          Player[i]^.GesamtPunkte := Player[i]^.GesamtPunkte+Diff;
          Player[j]^.GesamtPunkte := Player[j]^.GesamtPunkte-Diff;
        end;
      end;
      GetMem(a, 4096);
      StrPCopy(a, 'Punktestand:'+#13#10);
      for i := 0 to 3 do
        StrPCat(a, #13#10+FormatN(Player[i]^.GesamtPunkte,8)+' P  '+Player[i]^.name);

      MessageBox(Application^.MainWindow^.HWindow, a, 'GESAMTPUNKTE', mb_ok);
      FreeMem(a, 4096);

      {Eintragen in Highscore-Liste}
      SetHighscore;

      if Winner=East then
        EastCount := EastCount+1;
      if (EastCount=4) or (Winner<>East) then
      begin
        EastCount := 0;
        East := (East+1) mod 4;
      end;
      CalcPt := 0;
    end
    else
      NextGame;
    exit;
  end;

  if Austeilen>0 then
  begin
    repeat
      if Austeilen>1 then
      begin
        Player[Austeilen2]^.DrawTile(Wall.DrawLivingTile(Austeilen2));
        if (Austeilen2=East) or (Austeilen>2) then
          Player[Austeilen2]^.DrawTile(Wall.DrawLivingTile(Austeilen2));

        Austeilen2 := (Austeilen2+1) mod 4;
        if Austeilen2 = East then
          AusTeilen := AusTeilen-1;

        if Austeilen=1 then
        begin
          for i := 0 to 3 do
            Player[i]^.Hand.SortTiles;
          Player[0]^.Changed := True;
        end;
      end
      else
      begin
        i := 0;
        for j := East to East+3 do
          if i=0 then
          begin
            with Player[j mod 4]^ do
            begin
              T := Hand.GetTile(0);
              if T<8 then
              begin
                i := 1;
                Hand.RemoveTile(0);
                DrawBonusTile(T);
                InfoTxt := name+' zieht nach fr '+TileName[T];
                T := Wall.DrawDeadTile(0);
                DrawTile(T);
                Hand.SortTiles;
              end;
            end;
          end;
        if i=0 then
        begin
          Austeilen := 0;
          if East=0 then
          begin
            Beep('SPIELERZUG');
            InfoTxt := 'Dein Zug';
          end
          else
            WaitForWeiter := True;
        end;
      end;

    until (Austeilen=0) or AusteilenAn;
    for i := 0 to 3 do
      Player[i]^.SelectTaktik;
    exit;
  end;

  DoNotThrow := False;
  if WaitForCall then
  begin
    WaitForCall := False;
    PNr := -1;
    MaxCall := 0;
    T := Open.GetTile(Open.Count-1);
    for i := CurrentPlayer+1 to CurrentPlayer+3 do
    begin
      CL := Player[i mod 4]^.Call(T);
      if (CL=cl_Tschi) and (i<>CurrentPlayer+1) then
      begin
        Player[i mod 4]^.ThisCall := -1;
        CL := -1;
      end;
      if CL>MaxCall then
      begin
        MaxCall := CL;
        PNr := i mod 4;
      end;
      if CL=-1 then
        WaitForCall := True;
    end;
    if (not WaitForCall) then
    begin
      if PNr=-1 then
      begin
        CurrentPlayer := (CurrentPlayer+1) mod 4;
        if (Wall.Dead div 2) - (Wall.Living div 2)<=7 then
        begin
          finished := True;
          Beep('UNENTSCHIEDEN');
          InfoTxt := 'UNENTSCHIEDEN';
          exit;
        end;
      end
      else
      begin
        T := Open.GetTile(Open.Count-1);
        Open.RemoveTile(Open.Count-1);
        Player[PNr]^.ThisCall := -1;
        if Player[PNr]^.CalledTile(T, MaxCall) then
        begin
          CurrentPlayer := PNr;
          Player[CurrentPlayer]^.NextThrow := -1;
          if MaxCall=cl_kan then
          begin
            T := Wall.DrawDeadTile(0);
            Player[CurrentPlayer]^.DrawTile(T);
            DoNotThrow := True;
          end;
          if MaxCall=cl_MahJongg then
          begin
            Winner := CurrentPlayer;
            Beep('MAHJONGG');
            InfoTxt := Player[Winner]^.name+' hat gewonnen!';
            for i := 0 to 3 do
              Player[i]^.OpenHand;
            CalcPt := 1;
            Finished := True;
            exit;
          end;
          WaitForThrow := True;
          DoNotThrow := True;
          if CurrentPlayer=0 then
          begin
            Beep('SPIELERZUG');
            InfoTxt := 'Dein Zug';
          end;
        end
        else
        begin
          Open.AddTile(T);
          WaitForCall := True;
        end;
      end;
    end;
  end;


  if (not WaitForThrow) and (not WaitForCall) then 
  begin
    T := Wall.DrawLivingTile(CurrentPlayer);
    Player[CurrentPlayer]^.DrawTile(T);
    Player[CurrentPlayer]^.NextThrow := -1;
    WaitForThrow := True;
    if CurrentPlayer=0 then
    begin
      Beep('SPIELERZUG');
      InfoTxt := 'Dein Zug';
    end;
  end
  else if (not DoNotThrow) and WaitForThrow then
  begin
    CL := Player[CurrentPlayer]^.Call(NoTile);
    if CL = cl_Kan then
    begin
      if Player[CurrentPlayer]^.CalledTile(NoTile, CL) then
      begin
        T := Wall.DrawDeadTile(0);
        Player[CurrentPlayer]^.DrawTile(T);
        Player[CurrentPlayer]^.NextThrow := -1;
        Player[CurrentPlayer]^.ThisCall := 0;
        WaitForThrow := True;
        DoNotThrow := True;
      end;
    end
    else
    if CL=cl_MahJongg then
    begin
      if Player[CurrentPlayer]^.DoMahJongg(NoTile) then
      begin
        Winner := CurrentPlayer;
        Beep('MAHJONGG');
        InfoTxt := Player[Winner]^.name+' hat gewonnen!';
        for i := 0 to 3 do
          Player[i]^.OpenHand;
        CalcPt := 1;
        Finished := True;
        exit;
      end;
    end;
    if not DoNotThrow then
    begin
      if (CurrentPlayer=0) and (Player[0]^.NextThrow<>-1) then
        SaveGame;
      T := Player[CurrentPlayer]^.ThrowTile;
      if T<>NoTile then
      begin
        if (T<8) then
        begin
          Player[CurrentPlayer]^.DrawBonusTile(T);
          T := Wall.DrawDeadTile(0);
          Player[CurrentPlayer]^.DrawTile(T);
          Player[CurrentPlayer]^.NextThrow := -1;
          WaitForThrow := True;
        end
        else
        begin
          Open.AddTile(T);
          for i := CurrentPlayer+1 to CurrentPlayer+3 do
            Player[i mod 4]^.ThisCall := -1;
          Player[CurrentPlayer]^.ThisCall := 0;
          WaitForThrow := False;
          WaitForCall := True;
        end;
      end;
    end;
  end;
end;

procedure TGameClass.SaveGame;
 var
    i, j: Integer;
begin
  for i := 0 to 3 do
  begin
    Save.Player[i]^.Reset;
    for j := 0 to Player[i]^.Hand.Count-1 do
      Save.Player[i]^.Hand.AddTile(Player[i]^.Hand.GetTile(j));
    for j := 0 to Player[i]^.Open.Count-1 do
      Save.Player[i]^.Open.AddTile(Player[i]^.Open.GetTile(j));
  end;
  Save.Living := Wall.Living;
  Save.Dead := Wall.Dead;
  Save.OpenCount := Open.Count;
  Save.RSeed := RandSeed;
  Save.Valid := True;
end;

procedure TGameClass.LoadGame;
 var
    i, j: Integer;
begin
  if Save.Valid then
  begin
    for i := 0 to 3 do
    begin
      Player[i]^.Reset;
      for j := 0 to Save.Player[i]^.Hand.Count-1 do
        Player[i]^.Hand.AddTile(Save.Player[i]^.Hand.GetTile(j));
      for j := 0 to Save.Player[i]^.Open.Count-1 do
        Player[i]^.Open.AddTile(Save.Player[i]^.Open.GetTile(j));
      Player[i]^.Changed := True;
    end;
    Wall.Living := Save.Living;
    Wall.Dead := Save.Dead;
    Open.Count := Save.OpenCount;
    RandSeed := Save.RSeed;
    Player[0]^.ThisCall := cl_Nothing;
    Player[0]^.NextThrow := -1;
    CurrentPlayer := 0;
    WaitForThrow := True;
    WaitForCall := False;
    Finished := False;
    InfoTxt := 'Alter Spielstand wiederhergestellt';
    Beep('SPIELERZUG')
  end;
end;

function TGameClass.GetWallHeight(Nr, Pos: integer): Integer; 
 var
    LogNr, Height, LD2, DD2: Integer;
begin
  Height := 0;
  LogNr := (72-East*18+Nr*18+Pos+72-StartPos) mod 72;
  LD2 := Wall.Living div 2;
  DD2 := Wall.Dead div 2;
  if LogNr<LD2 then
    Height := 0
  else if LogNr=LD2 then
    Height := 2-Wall.Living mod 2
  else if LogNr>=DD2 then
    Height := 0
  else
  begin
    Height := 2;
    if LogNr = DD2-4 then
      Height := 3
    else if (LogNr = DD2-2) and ((Wall.Dead Mod 2)=1) then
      Height := 3;
  end;
  GetWallHeight := Height;
end;


function TGameClass.CountOpenTile(TName: Str2): Integer;
 var
    Sum, i: Integer;
begin
  Sum := Open.CountOpen(TName);
  for i := 0 to 3 do
    Sum := Sum+Player[i]^.Open.CountTile(TName);
  CountOpenTile := Sum;
end;



procedure TGameClass.ShowHighscore;
 var
    i: Integer;
    A: PChar;
    s2: String;
begin
  GetMem(a, 4096);
  StrPCopy(a, ' Punkte     Name'+#13#10+
              '______________');

  for i := 1 to 10 do
  begin
    s2 := FormatN(HighScore[i].Punkte, 8);
    if HighScore[i].Current then
      s2 := s2+' * '
    else
      s2 := s2+'    ';
    StrPCat(a, #13#10+s2+HighScore[i].Name);
  end;

  MessageBox(Application^.Mainwindow^.HWindow, a, 'HIGHSCORE', mb_ok);
  FreeMem(a, 4096);
end;


procedure TGameClass.SetHighscore;
 var
    k, i, j: Integer;
    NewHighScore: Boolean;
begin
  if Player[0]^.Name='Mogel' then
    exit;

  {Bestpunkte setzen}
  for i := 0 to 3 do
    if Player[i]^.GesamtPunkte>Player[i]^.BestPunkte then
      Player[i]^.BestPunkte := Player[i]^.GesamtPunkte;

  {Alle laufenden Spieler entfernen}
  for i := 10 downto 1 do
    if HighScore[i].Current then
    begin
      for j := i to 9 do
        HighScore[j] := HighScore[j+1];
      HighScore[10].Punkte := 0;
    end;

  {Alle laufenden Spieler eintragen}
  for i := 0 to 3 do
    if Player[i]^.BestPunkte>HighScore[10].Punkte then
    begin
      j := 10;
      while (j>1) and (HighScore[j-1].Punkte<Player[i]^.BestPunkte) do
        j := j-1;
      for k := 9 downto j do
        HighScore[k+1] := HighScore[k];
      HighScore[j].Name := Player[i]^.Name;
      HighScore[j].Punkte := Player[i]^.BestPunkte;
      HighScore[j].Spiele := AnzSpiele;
      HighScore[j].Current := True;
    end;

  ShowHighScore;
end;


procedure TGameClass.Reset;
 var
    i: integer;
begin
  Wall.Reset;
  Open.Reset;
  for i := 0 to 3 do
  begin
    Player[i]^.Reset;
    Player[i]^.BestPunkte := 0;
  end;
  StartPos := 17-random(5)-random(4)-3;
  for i := 1 to 10 do
    Highscore[i].Current := False;
  AnzSpiele := 0;
  East := random(4);
  EastCount := 0;
  CurrentPlayer := East;
  CalcPt := 0;
  NextGame;
  WaitForWeiter := False;
end;

destructor TGameClass.Done; 
 var
    i: Integer;
begin
  for i := 0 to 3 do
  begin
    Dispose(Player[i], Done);
    Dispose(Save.Player[i], Done);
  end;
  Wall.Done;
  Open.Done;
end;


var
  SaveExit: Pointer;

procedure MahJonggExit;
far;
begin
  ExitProc := SaveExit;
  TheGame.Done;
end;

begin
  TheGame.Init;
  SaveExit := ExitProc;
  ExitProc := @MahJonggExit;
end.

