unit mjDraw;

interface

uses mahjongg, Strings, WinProcs, WinTypes, GrTools;

var
   OpenTileBMP,
   VOpenTileBMP,
   VOTileMaskBMP,
   HVOpenTileBMP,
   HVOTileMaskBMP,
   OTileMaskBMP,
   TileFrameBMP,
   TileMaskBMP,
   ScheibenMaskBMP,
   HorzBackBMP,
   HBackMaskBMP,
   VertBackBMP,
   VBackMaskBMP: HBitmap;
   Scheiben: array[0..3, 0..1] of HBitmap;
   TileBMP: array[0..41] of record
                              Name: String[2];
                              BMP: HBitmap;
                            end;  
   TileDC: HDC;
   OldTileDCBmp: HBitmap;

procedure DrawTile(ADC: HDC; X, Y: Integer; T: Tile);
procedure DrawOpenTile(ADC: HDC; X, Y: Integer; T: Tile);
procedure DrawHVOpenTile(ADC: HDC; X, Y: Integer);
procedure DrawVOpenTile(ADC: HDC; X, Y: Integer; T: Tile);
procedure DrawHBack(ADC: HDC; X, Y: Integer);
procedure DrawVBack(ADC: HDC; X, Y: Integer);
procedure DrawScheibe(ADC: HDC; X, Y, R: Integer; Inverted: Boolean);

implementation


procedure DrawHBack(ADC: HDC; X, Y: Integer);
begin
  SelectObject(TileDC, HBackMaskBmp);
  BitBlt(ADC, X, Y, 32, 26, TileDC, 0, 0, MergePaint);
  SelectObject(TileDC, HorzBackBmp);
  BitBlt(ADC, X, Y, 32, 26, TileDC, 0, 0, SrcAnd);
end;


procedure DrawVBack(ADC: HDC; X, Y: Integer);
begin
  SelectObject(TileDC, VBackMaskBmp);
  BitBlt(ADC, X, Y, 26, 32, TileDC, 0, 0, MergePaint);
  SelectObject(TileDC, VertBackBmp);
  BitBlt(ADC, X, Y, 26, 32, TileDC, 0, 0, SrcAnd);
end;



procedure DrawTile(ADC: HDC; X, Y: Integer; T: Tile);
 var
    i, j: Integer;
    Name: String[2];
    r: TRect;
begin
  Name := TileName[T];
  j := 0;
  for i := 0 to 41 do
    if (TileBmp[i].Name=Name) then
      j := i;
  SelectObject(TileDC, TileMaskBmp);
  BitBlt(ADC, X, Y, 26, 32, TileDC, 0, 0, MergePaint);
  SelectObject(TileDC, TileFrameBmp);
  BitBlt(ADC, X, Y, 26, 32, TileDC, 0, 0, SrcAnd);
  SelectObject(TileDC, TileBmp[j].Bmp);
  BitBlt(ADC, X+2, Y+5, 19, 25, TileDC, 0, 0, SrcCopy);
end;

procedure DrawOpenTile(ADC: HDC; X, Y: Integer; T: Tile);
 var
    i, j: Integer;
    Name: String[2];
    r: TRect;
begin
  Name := TileName[T];
  j := 0;
  for i := 0 to 41 do
    if (TileBmp[i].Name=Name) then
      j := i;
  SelectObject(TileDC, OTileMaskBmp);
  BitBlt(ADC, X, Y, 23, 26, TileDC, 0, 0, MergePaint);
  SelectObject(TileDC, OpenTileBmp);
  BitBlt(ADC, X, Y, 23, 26, TileDC, 0, 0, SrcAnd);
  SelectObject(TileDC, TileBmp[j].Bmp);
  SetStretchBltMode(ADC, 4);  {HALFTONE}
  StretchBlt(ADC, X+3, Y+2, 17, 10, TileDC, 0,  0, 19, 15, SrcCopy);
  StretchBlt(ADC, X+2, Y+12, 19,  7, TileDC, 0, 15, 19, 10, SrcCopy);
end;

procedure DrawVOpenTile(ADC: HDC; X, Y: Integer; T: Tile);
 var
    i, j: Integer;
    Name: String[2];
    r: TRect;
begin
  Name := TileName[T];
  j := 0;
  for i := 0 to 41 do
    if (TileBmp[i].Name=Name) then
      j := i;
  SelectObject(TileDC, VOTileMaskBmp);
  BitBlt(ADC, X, Y, 26, 23, TileDC, 0, 0, MergePaint);
  SelectObject(TileDC, VOpenTileBmp);
  BitBlt(ADC, X, Y, 26, 23, TileDC, 0, 0, SrcAnd);
  SelectObject(TileDC, TileBmp[j].Bmp);
  for i := 0 to 4 do
    for j := 0 to 17 do
      SetPixel(ADC, x+j+4, y+i+5, GetPixel(TileDC, trunc(8.0*i/4.0), 24-trunc(24.0*j/17.0)));
  for i := 0 to 5 do
    for j := 0 to 19 do
      SetPixel(ADC, x+j+3, y+i+10, GetPixel(TileDC, 9+trunc(9.0*i/5.0), 24-trunc(24.0*j/19.0)));
end;

procedure DrawHVOpenTile(ADC: HDC; X, Y: Integer);
 var
    i, j: Integer;
    Name: String[2];
    r: TRect;
begin
  SelectObject(TileDC, HVOTileMaskBmp);
  BitBlt(ADC, X, Y, 23, 26, TileDC, 0, 0, MergePaint);
  SelectObject(TileDC, HVOpenTileBmp);
  BitBlt(ADC, X, Y, 23, 26, TileDC, 0, 0, SrcAnd);
end;

procedure DrawScheibe(ADC: HDC; X, Y, R: Integer; Inverted: Boolean);
begin
  SelectObject(TileDC, ScheibenMaskBmp);
  BitBlt(ADC, X, Y, 20, 20, TileDC, 0, 0, MergePaint);
  if Inverted then
    SelectObject(TileDC, Scheiben[R, 1])
  else
    SelectObject(TileDC, Scheiben[R, 0]);
  BitBlt(ADC, X, Y, 20, 20, TileDC, 0, 0, SrcAnd);
end;




var
  SaveExit: Pointer;



procedure mjDrawInit;
 var
    i, j: Integer;
    LastName: String[2];
    s: String;
    a: array[0..80] of char;
    ADC: HDC;
begin
  OpenTileBMP   := LoadBitmap(HInstance, 'BITMAP_OPENTILE');
  OTileMaskBMP  := LoadBitmap(HInstance, 'BITMAP_OTILEMASK');
  HVOpenTileBMP := LoadBitmap(HInstance, 'BITMAP_HVOPENTILE');
  HVOTileMaskBMP:= LoadBitmap(HInstance, 'BITMAP_HVOTILEMASK');
  VOpenTileBMP  := LoadBitmap(HInstance, 'BITMAP_VOPENTILE');
  VOTileMaskBMP := LoadBitmap(HInstance, 'BITMAP_VOTILEMASK');
  TileFrameBMP  := LoadBitmap(HInstance, 'BITMAP_FRAME');
  TileMaskBMP   := LoadBitmap(HInstance, 'BITMAP_TILEMASK');
  HBackMaskBMP  := LoadBitmap(HInstance, 'BITMAP_HBACKMASK');
  HorzBackBMP   := LoadBitmap(HInstance, 'BITMAP_HORZBACK');
  VBackMaskBMP  := LoadBitmap(HInstance, 'BITMAP_VBACKMASK');
  VertBackBMP   := LoadBitmap(HInstance, 'BITMAP_VERTBACK');

  ScheibenMaskBMP:=LoadBitmap(HInstance, 'BITMAP_SCHEIBE_MASK');
  Scheiben[0,0] := LoadBitmap(HInstance, 'BITMAP_SCHEIBE_E1');
  Scheiben[0,1] := LoadBitmap(HInstance, 'BITMAP_SCHEIBE_E2');
  Scheiben[1,0] := LoadBitmap(HInstance, 'BITMAP_SCHEIBE_S1');
  Scheiben[1,1] := LoadBitmap(HInstance, 'BITMAP_SCHEIBE_S2');
  Scheiben[2,0] := LoadBitmap(HInstance, 'BITMAP_SCHEIBE_W1');
  Scheiben[2,1] := LoadBitmap(HInstance, 'BITMAP_SCHEIBE_W2');
  Scheiben[3,0] := LoadBitmap(HInstance, 'BITMAP_SCHEIBE_N1');
  Scheiben[3,1] := LoadBitmap(HInstance, 'BITMAP_SCHEIBE_N2');

  LastName := '??';
  j := 0;
  for i:=0 to 143 do
    if TileName[i]<>LastName then
    begin
      LastName := TileName[i];
      s := 'BITMAP_'+LastName;
      StrPCopy(a, s);
      TileBmp[j].Name := LastName;
      TileBmp[j].Bmp := LoadBitmap(HInstance, a); 
      j := j+1;
    end;
{  Message(NToS(j));}

  ADC := GetDC(0);
  TileDC := CreateCompatibleDC(ADC);
  ReleaseDC(0, ADC);
  OldTileDCBmp := SelectObject(TileDC, TileBmp[0].Bmp);
end;


procedure mjDrawExit;
far;
 var
    i, j: integer;
begin
  ExitProc := SaveExit;

  SelectObject(TileDC, OldTileDCBmp);
  DeleteDC(TileDC);
  for i := 0 to 41 do
    DeleteObject(TileBmp[i].Bmp);

  for i := 0 to 3 do
    for j := 0 to 1 do
      DeleteObject(Scheiben[i,j]);

  DeleteObject(ScheibenMaskBMP);
  DeleteObject(OpenTileBMP);
  DeleteObject(OTileMaskBMP);
  DeleteObject(VOpenTileBMP);
  DeleteObject(VOTileMaskBMP);
  DeleteObject(HVOpenTileBMP);
  DeleteObject(HVOTileMaskBMP);
  DeleteObject(TileMaskBMP);
  DeleteObject(TileFrameBMP);
  DeleteObject(HorzBackBMP);
  DeleteObject(HBackMaskBMP);
  DeleteObject(VertBackBMP);
  DeleteObject(VBackMaskBMP);
end;

begin
  mjDrawInit;
  SaveExit := ExitProc;
  ExitProc := @mjDrawExit;
end.

