unit mjWindow;

Interface

uses mahJongg, Objects, WinProcs, WinTypes, mjDraw, GrTools, mjDefs, mjDlgs, OWindows;


type
  PPlayerWindow = ^TPlayerWindowClass;

  TPlayerWindowClass = object(TWindow)
    ToolsActive: Boolean;
    Player: PPlayer;
    Icons: array[0..6] of HIcon;
    IconsPerRow,
    StartRow: Integer;

    { Creation and destruction }
    constructor Init(AParent: PWindowsObject; APlayer: PPlayer);
    destructor Done; virtual;

    { Display }
    procedure Paint(PaintDC: HDC; var PaintInfo: TPaintStruct); virtual;

    { Window manager responses }

    procedure WMLButtonDown(var Msg: TMessage);
      virtual wm_First + wm_LButtonDown;

    procedure RedrawIfChanged; virtual;
    procedure ToolSelect(n: Integer); virtual;
  end;



implementation


{-------------- TPLAYERWINDOW -----------------}

constructor TPlayerWindowClass.Init(AParent: PWindowsObject; APlayer: PPlayer);
begin
  TWindow.Init(AParent, nil);
  Attr.Style := ws_PopUp or ws_Visible or ws_caption;
  Attr.X := trunc(135*1.6);
  Attr.Y := trunc(280*1.6);
  Attr.W := trunc(375*1.6);
  Attr.H := trunc((100 +30)*1.6);
  Player := APlayer;
  Icons[0] := LoadIcon(HInstance, 'ICON_WARTEN');
  Icons[1] := LoadIcon(HInstance, 'ICON_WEITER');
  Icons[2] := LoadIcon(HInstance, 'ICON_TSCHI');
  Icons[3] := LoadIcon(HInstance, 'ICON_PONG');
  Icons[4] := LoadIcon(HInstance, 'ICON_KAN');
  Icons[5] := LoadIcon(HInstance, 'ICON_MAHJONGG');
  Icons[6] := LoadIcon(HInstance, 'ICON_ENDE');
  ToolsActive := False;
end;

procedure TPlayerWindowClass.RedrawIfChanged;
begin
  if Player^.Changed then
  begin
    Player^.Changed := False;
    InvalidateRect(HWindow, nil, True);
  end;
end;



procedure TPlayerWindowClass.Paint(PaintDC: HDC; var PaintInfo: TPaintStruct);
var
  i: Integer;
  XPos, YPos: Integer;
  MaxX: Integer;
  Mode: Integer;

  width: Integer;
  height: Integer;
  MemDC: HDC;
  OldBmp: HBitmap;
begin
  MaxX := 7;   {mindestbreite von 7 Steinen, damit Titel noch lesbar bleibt}
  width := trunc(Attr.W / 1.6);
  height := trunc(Attr.H / 1.6);
  IconsPerRow := width div 32;

 { TWindow.Paint(PaintDC, PaintInfo);}
  MemDC := CreateCompatibleDC(PaintDC);
  OldBMP := CreateCompatibleBitmap(PaintDC, width, height);
  OldBMP := SelectObject(MemDC, OldBMP);

  BitBlt(MemDC, 0, 0, width, height, MemDC, 0, 0, Whiteness);

  for i := 0 to 6 do
    DrawIcon(MemDC, (i mod IconsPerRow)*32, (i div IconsPerRow)*32, Icons[i]);
  YPos := (1+(6 div IconsPerRow))*32-1;
  MoveTo(MemDC, 0, YPos);
  LineTo(MemDC, width, YPos);

  XPos := 0;
  YPos := YPos+9;
  Mode := NoTile;
  for i := 0 to Player^.Open.Count-1 do
  begin
    if Player^.Open.GetTile(i)=NoTile then
    begin
      Mode := NoTile;
      if XPos>0 then
      begin
        XPos := 0;
        YPos := YPos+28;
      end;
    end
    else
    if Player^.Open.GetTile(i)=HalbVerdeckt then
      Mode := HalbVerdeckt
    else
    if Player^.Open.GetTile(i)=Verdeckt then
      Mode := Verdeckt
    else
    begin
      MaxX := IMax(MaxX, XPos);
      if (Mode=Verdeckt) then
        DrawVOpenTile(MemDC, 5+XPos*26, YPos+3, Player^.Open.GetTile(i))
      else if (Mode=HalbVerdeckt) then
        DrawHVOpenTile(MemDC, 5+XPos*26, YPos)
      else
        DrawOpenTile(MemDC, 5+XPos*26, YPos, Player^.Open.GetTile(i));
      Mode := NoTile;
      XPos := XPos+1;
    end;
  end;

  if XPos = 0 then
    YPos := YPos-28;
  StartRow := YPos+30;

  XPos := 0;
  YPos := StartRow;
  for i := 0 to Player^.Hand.Count-1 do
  begin
    MaxX := IMax(MaxX, XPos);
    DrawTile(MemDC, 5+XPos*26, YPos, Player^.Hand.GetTile(i));
    XPos := XPos+1;
  end;

  SetStretchBltMode(PaintDC, 4);  {HALFTONE}
  StretchBlt(PaintDC, 0, 0, trunc(width*1.6), trunc(height*1.6), MemDC, 0, 0, width, height, SrcCopy);

  DeleteObject(SelectObject(MemDC, OldBMP));
  DeleteDC(MemDC);


  if XPos=0 then
    YPos := YPos-33;
  if (YPos+60<>height) or (width<45+MaxX*26) or (width-33>45+MaxX*26) then
    MoveWindow(HWindow, Attr.X, Attr.Y, trunc((45+MaxX*26)*1.6), trunc((YPos+60)*1.6), True);
end;



procedure TPlayerWindowClass.WMLButtonDown(var Msg: TMessage);
var
  x, y: Integer;
  r: TRect;
  ADC: HDC;
begin
  if not ToolsActive then
  begin
    ToolsActive := True;
    x := Integer(Msg.LParamLo);
    y := Integer(Msg.LParamHi);
    x := trunc(x/1.6);
    y := trunc(y/1.6);
    if (y>=0) and (y<=(1+(6 div IconsPerRow))*32-1) then
    begin
      y := y div 32;
      x := x div 32;
      if (y*IconsPerRow+x>=0) and (y*IconsPerRow+x<=6) and
         (x<IconsPerRow) then
      begin
        ADC := GetDC(HWindow);
        SetRect(r, trunc(x*32*1.6), trunc(y*32*1.6), trunc((x*32+32)*1.6), trunc((y*32+32)*1.6));
        InvertRect(ADC, r);
        ToolSelect(y*IconsPerRow+x);
        InvertRect(ADC, r);
        ReleaseDC(HWindow, ADC);
      end;
    end
    else
    if (y>=StartRow) then
    begin
      y := (y-StartRow) div 33;
      x := (x-5) div 26;
      if (y=0) and (x>=0) and (x<Player^.Hand.Count) then
      begin
        if Player^.NextThrow = -1 then
        begin
          Player^.NextThrow := x;
          PostMessage(Parent^.HWindow, WM_Command, cm_NextMove, 0);
        end;
      end;
    end;
    ToolsActive := False;
  end;
{  SetActiveWindow(Parent^.HWindow);}
end;


procedure TPlayerWindowClass.ToolSelect(n: Integer);
 var
    Buffer: TWaitInfo;
    OldTA: Boolean;
begin
  OldTA := TimerActive;
  TimerActive := True;
  case n of
    0    : begin
             if Application^.ExecDialog(New(PWaitDialog, Init(@Buffer)))= ID_OK then
             begin
               if Buffer.n>0 then
               begin
                 Player^.ThisCall := Buffer.n;
                 PostMessage(Parent^.HWindow, WM_Command, cm_NextMove, 0);
               end;
             end;
           end;
    1..5 : begin
             if n=1 then
               TheGame.WaitForWeiter := False;
             Player^.ThisCall := n-1;
             PostMessage(Parent^.HWindow, WM_Command, cm_NextMove, 0);
           end;  
    6    : if AppMessageString('Wollen Sie das Programm tatschlich beenden?',
                               'Mah Jongg', mb_YesNo+mb_IconQuestion) = id_yes then
             PostMessage(Parent^.HWindow, WM_Command, cm_Quit, 0);
  end;
  TimerActive := OldTA;
end;


destructor TPlayerWindowClass.Done;
var
  i: Integer;
begin
  for i := 0 to 6 do
    DestroyIcon(Icons[i]);
  TWindow.Done;
end;


end.
